#!/usr/bin/env python
# -*- coding: utf-8 -*-

#	IMPORTS

from netCDF4 import Dataset
from datetime import datetime, timedelta
import cmap as ccm
import math
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy.ma as ma
import matplotlib.ticker as ticker
import matplotlib.dates as mdates
import sys
import io
from PIL import Image
import os, fnmatch


#----------	PLACE, TIME, MODEL	----------

#datum='2019-03-15'
datum=sys.argv[1]
cloudloc=u"Meteorological Institute of Ludwig-Maximilians-Universität (München, Germany; 48.148 N / 11.573 E): "
date='WRF-d03: '+datum[-2:]+'-'+datum[5:-3]+'-'+datum[:4]


#----------	GRIDPOINT	----------

y = 1 
leftBound=0 
rightBound=2
upperBound=30 #obere Grenze des "Plot-Gitters"
hohe = 30 #Anzahl an Höhen-Leveln die geplottet werden


#----------	HANDY FUNCTIONS 	----------

#Überprüfung ob Daten eine gültige Zahl enhalten oder nicht 
def isMasked(zahl):
    return isinstance(zahl, np.ma.core.MaskedConstant)

#Konvertiert einen Eintrag in den Daten in eine Gleitkommazahl, ist kein Eintrag Vorhand wird von 0 ausgegangen
def toFloat(zahl):
    if (isMasked(zahl)):
        return 0.0
    
    return zahl

#Berechnung der Höhe einer Zelle
def calcHeight(height, heightData):
    if height == 0:
        return heightData[height]/2.0
    else:
        return (heightData[height] - heightData[height-1])
#  this was not right.
#    if height == 0 or height == 1:
#        return heightData[height]/2.0
#    else:
#        return (heightData[height] - heightData[height-2])/2.0

#Berechnung der Minmalreflektivität in Abhängigkeit der Höhe
def calcMinReflectivity(curHeight):
    c0 = -26.1
    z0 = 5000
    snr_min = -18.6
    
    return c0 + 20*np.log10(curHeight/z0) + snr_min
  
  
#----------	READING IN THE DATA	----------

path_to_files = "/project/meteo/scratch/met-wrf-chem/"
	
#filenames = [path_to_files + "0" + str(H) + ":00:00.nc" for H in range(10)] + [path_to_files + str(H) + ":00:00.nc" for H in range(10, 24)]
filenames = fnmatch.filter(os.listdir(path_to_files), '??:??:??.nc')
filenames = [ path_to_files+filename for filename in filenames ]
filenames.sort()


#Resiervierung von Speicher für die Ergebnisse
dampfungsWerte = ma.array(np.ndarray(shape=(len(filenames),hohe)), mask=[False*3])

#Für jede Datei wird ein Array für München ausglesen
for filename in enumerate(filenames):
    werteIndex = filename[0]
    dataset = Dataset(filename[1])

    #Auslesen der Reflektiviätswerte aus den Datensatz für München (oder den gewünschten Standort)
    slice = dataset.variables['Zhh'][:upperBound,y,leftBound:rightBound] 
	
    #Berechnung der Dämpfungen in [dB/meter] 
    ah = ((dataset.variables['Ah'][:upperBound,y,leftBound:rightBound])/1000)
    
    #Höhendaten für den Plot
    heightData = dataset.variables['height'][:upperBound,y,leftBound:rightBound]

    #Tauschen der Achsen für die Vereinfachung der Rechnung
    slice = np.swapaxes(slice, 0, 1)

    #Für jeden Eintrag im Array wird nun die tatsächliche Reflektivität berechnet
    for x in range(len(slice)):
        hintererTerm = 0.0
        for height in range(len(slice[x])-1):
            if (isMasked(ah[height, x])):
                continue
            
            curAh = toFloat(ah[height, x])
            curHeight = toFloat(calcHeight(height, heightData[:,x]))

            hintererTerm += 2.0 * curAh * curHeight
            slice[x,height] = toFloat(slice[x,height]) - hintererTerm
            
            minReflectivity = calcMinReflectivity(heightData[height, x])
            
            if (slice[x, height] < minReflectivity):
                slice.mask[x, height] = True  


    #Zurücktauschen der Achsen für die korrekte Visualisierung der Daten
    slice = np.swapaxes(slice, 0, 1)

    dampfungsWerte[werteIndex]=slice[:hohe,0]
    dampfungsWerte.mask[werteIndex]=slice.mask[:hohe,0]


#----------	PREPARATION FOR THE PLOT	----------

sunrise,sunset=ccm.get_sunrise_and_sunset(datum)
#time_axis = [datetime(2019, 3, 15, x, 0, 0) for x in range(24)]
#t0=datetime.strptime(datum,'%y%m%d')

n_ts=len(filenames) #Anzahl von Messpunkten an einem Tag
ts=int(len(filenames)*3600/n_ts*1000)/1000.
timelines=[int(n_ts/12.*i) for i in range(13)]

vmin=-40 #minimaler Wert der Colorbar                                                                                                                                         
vmax=30 #maximaler Wert der Colorbar                    


#Achsen für die dampfungsWerte
dampfungsWerte = np.swapaxes(dampfungsWerte, 0, 1)

#SNR_masked,x_nodata,dts=ccm.create_const_time_grid(dampfungsWerte,time_axis,t0,ts,n_ts,factor=np.nan) #not using SNR_masked, just using x_nodata
#x_nodata = ccm.create_const_time_grid(dampfungsWerte,time_axis,t0,ts,n_ts,factor=np.nan)[1]


#----------	PLOTTING	----------

colours={'dark_background':{'cb':'white','axes':'yellow'},'default':{'cb':'k','axes':'k'}}
style='default' #'default' für weißen Hintergrund auswählen
fs=38 #Fontsize fuer die Achsenbeschriftungen etc.
plt.style.use(style) 

fig,ax=plt.subplots(1,1,figsize=(85.12,18.98))
plot=plt.imshow(dampfungsWerte,cmap='jet',vmin=vmin,vmax=vmax,aspect='auto',origin='lower', extent=[-0.5, len(filenames)-0.5, 0, 14])
#plot=plt.imshow(dampfungsWerte,cmap='jet',vmin=vmin,vmax=vmax,aspect='auto',origin='lower')

#sunrise, sunset
ax.axvline(x=int(sunrise[0:2])*n_ts/24.+int(sunrise[2:4])*n_ts/24./60.,color=colours[style]['cb'],linestyle='--',linewidth=7) #Sonnenaufgang einzeichnen
ax.text(int(sunrise[0:2])*n_ts/24.+int(sunrise[2:4])*n_ts/24./60.,-1,'sunrise',fontsize=fs,horizontalalignment='center',color=colours[style]['axes'])

ax.axvline(x=int(sunset[0:2])*n_ts/24.+int(sunset[2:4])*n_ts/24./60.,color=colours[style]['cb'],linestyle='--',linewidth=7) #Sonnenuntergang einzeichnen
ax.text(int(sunset[0:2])*n_ts/24.+int(sunset[2:4])*n_ts/24./60.,-1,'sunset',fontsize=fs,horizontalalignment='center',color=colours[style]['axes'])

#title, axes, grid...
plt.title(cloudloc+"   "+date,fontsize=fs,color=colours[style]['axes'])
ax.set_yticks([0,2,4,6,8,10,12,14])
ax.set_ylim([0., 14.])
ax.set_xlim([0., 24.])
ax.set_yticklabels(['0.','2.','4.','6.','8.','10.','12.','14.'],color=colours[style]['axes'])
ax.set_xticks(timelines)
ax.set_xticklabels(['00','02','04','06','08','10','12','14','16','18','20','22','24'],color=colours[style]['axes'])
plt.tick_params(labelsize=fs,color=colours[style]['axes'])
ax.set_xlabel('Time [UTC]',fontsize=fs,color=colours[style]['axes'])
plt.ylabel('Height above ground [km]',fontsize=fs,color=colours[style]['axes'])
plt.grid(color=colours[style]['axes'],linewidth=2,linestyle='--')

#for i in x_nodata:
#    ax.axvline(x=i,color='lightgray',linewidth=1)

#colorbar
cb=plt.colorbar(plot,ticks=[-40,-30,-20,-10, 0, 10, 20, 30])
cb.ax.tick_params(labelsize=fs,color=colours[style]['cb'])
cb.set_label('Z [dBZ]',fontsize=fs,color=colours[style]['cb'])

#Save the plot at the desired location 
png_maindir='/project/meteo/homepages/quicklooks/wrf/reflectivity/' #Dort sollen die Plots grundsätzlich hin 
#png_maindir='/project/meteo/scratch/met-wrf-chem/cloudradar/'         
fig.savefig(png_maindir+'current.png', bbox_inches='tight', pad_inches = 0.1)





